# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.

from mock import patch, PropertyMock, MagicMock
from nose.tools import (
    raises,
    assert_equals,
    with_setup,
    assert_is_not_none,
    assert_false,
    assert_true,
)
import requests
from requests_kerberos.kerberos_ import HTTPKerberosAuth, REQUIRED, OPTIONAL
from sparkmagic.auth.basic import Basic
from sparkmagic.auth.kerberos import Kerberos
from sparkmagic.auth.customauth import Authenticator
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.livyclientlib.exceptions import HttpClientException
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException
from sparkmagic.livyclientlib.linearretrypolicy import LinearRetryPolicy
from sparkmagic.livyclientlib.reliablehttpclient import ReliableHttpClient
import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants

retry_policy = None
sequential_values = []
basic_auth = Basic()
kerberos_auth = Kerberos()
endpoint = Endpoint("http://url.com", basic_auth)


def _setup():
    global retry_policy
    retry_policy = LinearRetryPolicy(0.01, 5)


def _teardown():
    pass


def return_sequential():
    global sequential_values
    val = sequential_values[0]
    sequential_values = sequential_values[1:]
    return val


@with_setup(_setup, _teardown)
def test_compose_url():
    client = ReliableHttpClient(endpoint, {}, retry_policy)

    composed = client.compose_url("r")
    assert_equals("http://url.com/r", composed)

    composed = client.compose_url("/r")
    assert_equals("http://url.com/r", composed)

    client = ReliableHttpClient(endpoint, {}, retry_policy)

    composed = client.compose_url("r")
    assert_equals("http://url.com/r", composed)

    composed = client.compose_url("/r")
    assert_equals("http://url.com/r", composed)


@with_setup(_setup, _teardown)
def test_get():
    with patch("requests.Session.get") as patched_get:
        type(patched_get.return_value).status_code = 200

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.get("r", [200])

        assert_equals(200, result.status_code)


@raises(HttpClientException)
@with_setup(_setup, _teardown)
def test_get_throws():
    with patch("requests.Session.get") as patched_get:
        type(patched_get.return_value).status_code = 500

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        client.get("r", [200])


@with_setup(_setup, _teardown)
def test_get_will_retry():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = True
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.get") as patched_get:
        # When we call assert_equals in this unit test, the side_effect is executed.
        # So, the last status_code should be repeated.
        sequential_values = [500, 200, 200]
        pm = PropertyMock()
        pm.side_effect = return_sequential
        type(patched_get.return_value).status_code = pm
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.get("r", [200])

        assert_equals(200, result.status_code)
        retry_policy.should_retry.assert_called_once_with(500, False, 0)
        retry_policy.seconds_to_sleep.assert_called_once_with(0)


@with_setup(_setup, _teardown)
def test_post():
    with patch("requests.Session.post") as patched_post:
        type(patched_post.return_value).status_code = 200

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.post("r", [200], {})

        assert_equals(200, result.status_code)


@raises(HttpClientException)
@with_setup(_setup, _teardown)
def test_post_throws():
    with patch("requests.Session.post") as patched_post:
        type(patched_post.return_value).status_code = 500

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        client.post("r", [200], {})


@with_setup(_setup, _teardown)
def test_post_will_retry():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = True
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.post") as patched_post:
        # When we call assert_equals in this unit test, the side_effect is executed.
        # So, the last status_code should be repeated.
        sequential_values = [500, 200, 200]
        pm = PropertyMock()
        pm.side_effect = return_sequential
        type(patched_post.return_value).status_code = pm
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.post("r", [200], {})

        assert_equals(200, result.status_code)
        retry_policy.should_retry.assert_called_once_with(500, False, 0)
        retry_policy.seconds_to_sleep.assert_called_once_with(0)


@with_setup(_setup, _teardown)
def test_delete():
    with patch("requests.Session.delete") as patched_delete:
        type(patched_delete.return_value).status_code = 200

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.delete("r", [200])

        assert_equals(200, result.status_code)


@raises(HttpClientException)
@with_setup(_setup, _teardown)
def test_delete_throws():
    with patch("requests.Session.delete") as patched_delete:
        type(patched_delete.return_value).status_code = 500

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        client.delete("r", [200])


@with_setup(_setup, _teardown)
def test_delete_will_retry():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = True
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.delete") as patched_delete:
        # When we call assert_equals in this unit test, the side_effect is executed.
        # So, the last status_code should be repeated.
        sequential_values = [500, 200, 200]
        pm = PropertyMock()
        pm.side_effect = return_sequential
        type(patched_delete.return_value).status_code = pm
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.delete("r", [200])

        assert_equals(200, result.status_code)
        retry_policy.should_retry.assert_called_once_with(500, False, 0)
        retry_policy.seconds_to_sleep.assert_called_once_with(0)


@with_setup(_setup, _teardown)
def test_will_retry_error_no():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = False
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.get") as patched_get:
        patched_get.side_effect = requests.exceptions.ConnectionError()
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        try:
            client.get("r", [200])
            assert False
        except HttpClientException:
            retry_policy.should_retry.assert_called_once_with(None, True, 0)


@with_setup(_setup, _teardown)
def test_basic_auth_check_auth():
    endpoint = Endpoint("http://url.com", basic_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert isinstance(client._auth, Basic)
    assert hasattr(client._auth, "username")
    assert hasattr(client._auth, "password")
    assert_equals(client._auth.username, endpoint.auth.username)
    assert_equals(client._auth.password, endpoint.auth.password)


@with_setup(_setup, _teardown)
def test_no_auth_check_auth():
    endpoint = Endpoint("http://url.com", None)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert_equals(client._auth, None)


@with_setup(_setup, _teardown)
def test_kerberos_auth_check_auth():
    endpoint = Endpoint("http://url.com", kerberos_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert_is_not_none(client._auth)
    assert isinstance(client._auth, HTTPKerberosAuth)
    assert hasattr(client._auth, "mutual_authentication")
    assert_equals(client._auth.mutual_authentication, REQUIRED)


@with_setup(_setup, _teardown)
def test_kerberos_auth_custom_configuration():
    custom_kerberos_conf = {"mutual_authentication": OPTIONAL, "force_preemptive": True}
    overrides = {conf.kerberos_auth_configuration.__name__: custom_kerberos_conf}
    conf.override_all(overrides)
    kerberos_auth = Kerberos()
    endpoint = Endpoint("http://url.com", kerberos_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert_is_not_none(client._auth)
    assert isinstance(client._auth, HTTPKerberosAuth)
    assert hasattr(client._auth, "mutual_authentication")
    assert_equals(client._auth.mutual_authentication, OPTIONAL)
    assert hasattr(client._auth, "force_preemptive")
    assert_equals(client._auth.force_preemptive, True)



from mock import MagicMock
from nose.tools import assert_equals, assert_not_equals, raises, with_setup
import json

import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException
from sparkmagic.utils.constants import AUTH_BASIC, NO_AUTH


def _setup():
    conf.override_all({})


@with_setup(_setup)
def test_configuration_override_base64_password():
    kpc = {
        "username": "U",
        "password": "P",
        "base64_password": "cGFzc3dvcmQ=",
        "url": "L",
        "auth": AUTH_BASIC,
    }
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    assert_equals(
        conf.d,
        {
            conf.kernel_python_credentials.__name__: kpc,
            conf.livy_session_startup_timeout_seconds.__name__: 1,
        },
    )
    assert_equals(conf.livy_session_startup_timeout_seconds(), 1)
    assert_equals(
        conf.base64_kernel_python_credentials(),
        {"username": "U", "password": "password", "url": "L", "auth": AUTH_BASIC},
    )


@with_setup(_setup)
def test_configuration_auth_missing_basic_auth():
    kpc = {"username": "U", "password": "P", "url": "L"}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    assert_equals(
        conf.base64_kernel_python_credentials(),
        {"username": "U", "password": "P", "url": "L", "auth": AUTH_BASIC},
    )


@with_setup(_setup)
def test_configuration_auth_missing_no_auth():
    kpc = {"username": "", "password": "", "url": "L"}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    assert_equals(
        conf.base64_kernel_python_credentials(),
        {"username": "", "password": "", "url": "L", "auth": NO_AUTH},
    )


@with_setup(_setup)
def test_configuration_override_fallback_to_password():
    kpc = {"username": "U", "password": "P", "url": "L", "auth": NO_AUTH}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    assert_equals(
        conf.d,
        {
            conf.kernel_python_credentials.__name__: kpc,
            conf.livy_session_startup_timeout_seconds.__name__: 1,
        },
    )
    assert_equals(conf.livy_session_startup_timeout_seconds(), 1)
    assert_equals(conf.base64_kernel_python_credentials(), kpc)


@with_setup(_setup)
def test_configuration_override_work_with_empty_password():
    kpc = {
        "username": "U",
        "base64_password": "",
        "password": "",
        "url": "",
        "auth": AUTH_BASIC,
    }
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    assert_equals(
        conf.d,
        {
            conf.kernel_python_credentials.__name__: kpc,
            conf.livy_session_startup_timeout_seconds.__name__: 1,
        },
    )
    assert_equals(conf.livy_session_startup_timeout_seconds(), 1)
    assert_equals(
        conf.base64_kernel_python_credentials(),
        {"username": "U", "password": "", "url": "", "auth": AUTH_BASIC},
    )


@raises(BadUserConfigurationException)
@with_setup(_setup)
def test_configuration_raise_error_for_bad_base64_password():
    kpc = {"username": "U", "base64_password": "P", "url": "L"}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    conf.base64_kernel_python_credentials()


@with_setup(_setup)
def test_share_config_between_pyspark_and_pyspark3():
    kpc = {
        "username": "U",
        "password": "P",
        "base64_password": "cGFzc3dvcmQ=",
        "url": "L",
        "auth": AUTH_BASIC,
    }
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    assert_equals(
        conf.base64_kernel_python3_credentials(),
        conf.base64_kernel_python_credentials(),
    )



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
import asyncio
import inspect
import requests


from ipykernel.ipkernel import IPythonKernel
from hdijupyterutils.ipythondisplay import IpythonDisplay

import sparkmagic.utils.configuration as conf
from sparkmagic.utils.sparklogger import SparkLog
from sparkmagic.livyclientlib.exceptions import wrap_unexpected_exceptions
from sparkmagic.kernels.wrapperkernel.usercodeparser import UserCodeParser

# NOTE: This is a (hopefully) temporary workaround to accommodate async do_execute in ipykernel>=6
import nest_asyncio


# NOTE: This is a (hopefully) temporary workaround to accommodate async do_execute in ipykernel>=6
# Taken from: https://github.com/jupyter/notebook/blob/eb3a1c24839205afcef0ba65ace2309d38300a2b/notebook/utils.py#L332
def run_sync(maybe_async):
    """If async, runs maybe_async and blocks until it has executed,
    possibly creating an event loop.
    If not async, just returns maybe_async as it is the result of something
    that has already executed.
    Parameters
    ----------
    maybe_async : async or non-async object
        The object to be executed, if it is async.
    Returns
    -------
    result :
        Whatever the async object returns, or the object itself.
    """
    if not inspect.isawaitable(maybe_async):
        # that was not something async, just return it
        return maybe_async
    # it is async, we need to run it in an event loop

    def wrapped():
        create_new_event_loop = False
        result = None
        loop = None
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            create_new_event_loop = True
        else:
            if loop.is_closed():
                create_new_event_loop = True
        if create_new_event_loop:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        try:
            result = loop.run_until_complete(maybe_async)
        except RuntimeError as e:
            if str(e) == "This event loop is already running":
                # just return a Future, hoping that it will be awaited
                result = asyncio.ensure_future(maybe_async)
        return result

    return wrapped()


class SparkKernelBase(IPythonKernel):
    def __init__(
        self,
        implementation,
        implementation_version,
        language,
        language_version,
        language_info,
        session_language,
        user_code_parser=None,
        **kwargs
    ):
        # Required by Jupyter - Override
        self.implementation = implementation
        self.implementation_version = implementation_version
        self.language = language
        self.language_version = language_version
        self.language_info = language_info

        # Override
        self.session_language = session_language

        # NOTE: This is a (hopefully) temporary workaround to accommodate async do_execute in ipykernel>=6
        # Patch loop.run_until_complete as early as possible
        try:
            nest_asyncio.apply()
        except RuntimeError:
            # nest_asyncio requires a running loop in order to patch.
            # In tests the loop may not have been created yet.
            pass

        super().__init__(**kwargs)

        self.logger = SparkLog("{}_jupyter_kernel".format(self.session_language))
        self._fatal_error = None
        self.ipython_display = IpythonDisplay()

        if user_code_parser is None:
            self.user_code_parser = UserCodeParser()
        else:
            self.user_code_parser = user_code_parser

        # Disable warnings for test env in HDI
        requests.packages.urllib3.disable_warnings()

        # Do not load magics in testing
        if kwargs.get("testing", False):
            return

        # Load magics on init
        self._load_magics_extension()
        self._change_language()
        if conf.use_auto_viz():
            self._register_auto_viz()

    def do_execute(
        self, code, silent, store_history=True, user_expressions=None, allow_stdin=False
    ):
        def f(self):
            if self._fatal_error is not None:
                return self._repeat_fatal_error()

            return self._do_execute(
                code, silent, store_history, user_expressions, allow_stdin
            )

        # Execute the code and handle exceptions
        wrapped = wrap_unexpected_exceptions(f, self._complete_cell)
        return wrapped(self)

    def do_shutdown(self, restart):
        # Cleanup
        self._delete_session()

        return self._do_shutdown_ipykernel(restart)

    def _do_execute(self, code, silent, store_history, user_expressions, allow_stdin):
        code_to_run = self.user_code_parser.get_code_to_run(code)

        res = self._execute_cell(
            code_to_run, silent, store_history, user_expressions, allow_stdin
        )

        return res

    def _load_magics_extension(self):
        register_magics_code = "%load_ext sparkmagic.kernels"
        self._execute_cell(
            register_magics_code,
            True,
            False,
            shutdown_if_error=True,
            log_if_error="Failed to load the Spark kernels magics library.",
        )
        self.logger.debug("Loaded magics.")

    def _change_language(self):
        register_magics_code = "%%_do_not_call_change_language -l {}\n ".format(
            self.session_language
        )
        self._execute_cell(
            register_magics_code,
            True,
            False,
            shutdown_if_error=True,
            log_if_error="Failed to change language to {}.".format(
                self.session_language
            ),
        )
        self.logger.debug("Changed language.")

    def _register_auto_viz(self):
        from sparkmagic.utils.sparkevents import get_spark_events_handler
        import autovizwidget.utils.configuration as c

        handler = get_spark_events_handler()
        c.override("events_handler", handler)

        register_auto_viz_code = """from autovizwidget.widget.utils import display_dataframe
ip = get_ipython()
ip.display_formatter.ipython_display_formatter.for_type_by_name('pandas.core.frame', 'DataFrame', display_dataframe)"""
        self._execute_cell(
            register_auto_viz_code,
            True,
            False,
            shutdown_if_error=True,
            log_if_error="Failed to register auto viz for notebook.",
        )
        self.logger.debug("Registered auto viz.")

    def _delete_session(self):
        code = "%%_do_not_call_delete_session\n "
        self._execute_cell_for_user(code, True, False)

    def _execute_cell(
        self,
        code,
        silent,
        store_history=True,
        user_expressions=None,
        allow_stdin=False,
        shutdown_if_error=False,
        log_if_error=None,
    ):
        reply_content = self._execute_cell_for_user(
            code, silent, store_history, user_expressions, allow_stdin
        )

        if shutdown_if_error and reply_content["status"] == "error":
            error_from_reply = reply_content["evalue"]
            if log_if_error is not None:
                message = '{}\nException details:\n\t"{}"'.format(
                    log_if_error, error_from_reply
                )
                return self._abort_with_fatal_error(message)

        return reply_content

    def _execute_cell_for_user(
        self, code, silent, store_history=True, user_expressions=None, allow_stdin=False
    ):
        result = super().do_execute(
            code, silent, store_history, user_expressions, allow_stdin
        )

        # In ipykernel 6, this returns native asyncio coroutine
        if asyncio.iscoroutine(result):
            return run_sync(result)

        # In ipykernel 5, this returns gen.coroutine
        if asyncio.isfuture(result):
            return result.result()

        # In ipykernel 4, this func is synchronous
        return result

    def _do_shutdown_ipykernel(self, restart):
        result = super().do_shutdown(restart)

        # In tests, super() calls this SparkKernelBase.do_shutdown, which is async
        if asyncio.iscoroutine(result):
            return run_sync(result)

        return result

    def _complete_cell(self):
        """A method that runs a cell with no effect.

        Call this and return the value it returns when there's some sort
        of error preventing the user's cell from executing; this will
        register the cell from the Jupyter UI as being completed.
        """
        return self._execute_cell("None", False, True, None, False)

    def _show_user_error(self, message):
        self.logger.error(message)
        self.ipython_display.send_error(message)

    def _queue_fatal_error(self, message):
        """Queues up a fatal error to be thrown when the next cell is executed;
        does not raise an error immediately.

        We use this for errors that happen on kernel startup, since
        IPython crashes if we throw an exception in the __init__ method.
        """
        self._fatal_error = message

    def _abort_with_fatal_error(self, message):
        """Queues up a fatal error and throws it immediately."""
        self._queue_fatal_error(message)
        return self._repeat_fatal_error()

    def _repeat_fatal_error(self):
        """Throws an error that has already been queued."""
        error = conf.fatal_error_suggestion().format(self._fatal_error)
        self.logger.error(error)
        self.ipython_display.send_error(error)
        return self._complete_cell()



# coding=utf-8
from nose.tools import assert_equals

from sparkmagic.kernels.wrapperkernel.usercodeparser import UserCodeParser
from sparkmagic.kernels.kernelmagics import KernelMagics


def test_empty_string():
    parser = UserCodeParser()

    assert_equals("", parser.get_code_to_run(""))


def test_spark_code():
    parser = UserCodeParser()
    cell = "my code\nand more"

    assert_equals("%%spark\nmy code\nand more", parser.get_code_to_run(cell))


def test_local_single():
    parser = UserCodeParser()
    cell = """%local
hi
hi
hi"""

    assert_equals("hi\nhi\nhi", parser.get_code_to_run(cell))


def test_local_double():
    parser = UserCodeParser()
    cell = """%%local
hi
hi
hi"""

    assert_equals("hi\nhi\nhi", parser.get_code_to_run(cell))


def test_our_line_magics():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = "%{}".format(magic_name)

    assert_equals("%%{}\n ".format(magic_name), parser.get_code_to_run(cell))


def test_our_line_magics_with_content():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = """%{}
my content
more content""".format(
        magic_name
    )

    assert_equals(
        "%%{}\nmy content\nmore content\n ".format(magic_name),
        parser.get_code_to_run(cell),
    )


def test_other_cell_magic():
    parser = UserCodeParser()
    cell = """%%magic
hi
hi
hi"""

    assert_equals("{}".format(cell), parser.get_code_to_run(cell))


def test_other_line_magic():
    parser = UserCodeParser()
    cell = """%magic
hi
hi
hi"""

    assert_equals(cell, parser.get_code_to_run(cell))


def test_scala_code():
    parser = UserCodeParser()
    cell = """/* Place the cursor in the cell and press SHIFT + ENTER to run */

val fruits = sc.textFile("wasb:///example/data/fruits.txt")
val yellowThings = sc.textFile("wasb:///example/data/yellowthings.txt")"""

    assert_equals("%%spark\n{}".format(cell), parser.get_code_to_run(cell))


def test_unicode():
    parser = UserCodeParser()
    cell = "print 'è🐙🐙🐙🐙'"

    assert_equals("%%spark\n{}".format(cell), parser.get_code_to_run(cell))


def test_unicode_in_magics():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = """%{}
my content è🐙
more content""".format(
        magic_name
    )

    assert_equals(
        "%%{}\nmy content è🐙\nmore content\n ".format(magic_name),
        parser.get_code_to_run(cell),
    )


def test_unicode_in_double_magics():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = """%%{}
my content è🐙
more content""".format(
        magic_name
    )

    assert_equals(
        "%%{}\nmy content è🐙\nmore content\n ".format(magic_name),
        parser.get_code_to_run(cell),
    )



__version__ = "0.20.0"

from sparkmagic.serverextension.handlers import load_jupyter_server_extension


def _jupyter_server_extension_paths():
    return [{"module": "sparkmagic"}]


def _jupyter_nbextension_paths():
    return []
